#ifdef _WIN32
#include <windows.h>
#endif // _WIN32

#include <GL/gl.h>
#include <GL/glu.h>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <iostream>
#include <string>

#include <boost/shared_ptr.hpp>

#include <PinholeCamera.h>
#include <ErrorManager.h>
#include <GageAdaptor.h>
#include <IMarchingMethod.h>
#include <MarchingCubes.h>
#include <CombinedTransforms.h>
#include <VertexPool.h>
#include <Timer.h>
#include <MeshQuality.h>
#include <BoundingBox.h>

using namespace std;

// Image.
string m_imageName;
boost::shared_ptr<CGageAdaptor> m_image;
// Polygonizer.
boost::shared_ptr<IMarchingMethod> m_polygonizer;
float m_isoLevel = 0.0f;
bool m_drawEdges = false;
bool m_isUsingCombinedTransforms = false;
int m_transformationStepCount = 8;
CBoundingBox m_meshBoundingBox;
// Quality measurer.
boost::shared_ptr<CMeshQuality> m_qualityMeasurer;
// Reference grid.
unsigned int  m_gridDisplayList;
// Camera.
CPinholeCamera m_camera;
int m_mouseButton = 0;
int m_mouseX = -1;
int m_mouseY = -1;

/**
*/
bool ParseCommandLineOptions(int argc, char **argv)
{
	argv++;
	argc--;

	while (argc > 0)
	{
		if (strcmp(*argv, "-isCombined") == 0)
			m_isUsingCombinedTransforms = true;
		else if (strcmp(*argv, "-transformSteps") == 0)
		{
			if (argc > 1)
			{
				m_transformationStepCount = atoi(*(argv + 1));
				
				argv++;
				argc--;
			}
			else
			{
				MarkError();

				return false;
			}
		}

		argv++;
		argc--;
	}

	return true;
}

/**
*/
bool LoadImage(const string& path)
{
	double kernelParam[3];

	// Scale parameter, in units of samples.
    kernelParam[0] = 1.0; 
	// B.
    kernelParam[1] = 0.0;
	// C.
    kernelParam[2] = 0.5;

	m_image.reset(new CGageAdaptor);

	if (!m_image.get())
	{
		MarkError();

		return false;
	}

	if (!m_image->Open(path))
	{
		MarkError();

		return false;
	}

	// BC family of cubic polynomial splines (Catmull-Rom spline at 
	// (B,C)=(0,0.5)).
	if (!m_image->SetValueKernel(nrrdKernelBCCubic, kernelParam))
	{
		MarkError();

		return false;
	}

	// 1st deriv. of BC cubic family.
	if (!m_image->Set1stDerivativeKernel(nrrdKernelBCCubicD, kernelParam))
	{
		MarkError();

		return false;
	}

	m_imageName = path;

	return true;
}

/**
(test purposes only).
*/
bool MapQualityToColorArray(boost::shared_ptr<CVertexPool> mesh, boost::shared_ptr<CMeshQuality> qualityMeasurer)
{
	float colorTable[] = {
		0.0f, 0.0f, 1.0f, 
		1.0f, 0.0f, 1.0f, 
		1.0f, 0.5f, 0.0f, 
		1.0f, 1.0f, 0.0f, 
		0.0f, 1.0f, 0.0f, 
		0.0f, 1.0f, 0.0f
	};
	int firstColor,
		secondColor,
		i,
		j;
	float vertexQuality,
		mapping,
		t;

	mesh->m_colorArray.clear();
	mesh->m_colorSize = 3;

	mesh->m_colorArray.reserve(3*mesh->m_vertexArray.size());

	for (j=0; j!=(int)mesh->m_vertexArray.size()/mesh->m_vertexSize; j++)
	{
		vertexQuality = qualityMeasurer->GetVertexQuality(j);

		mapping = vertexQuality*5.0f;

		firstColor = (int)mapping;

		if (firstColor == 5)
		{
			t = 1.0f;

			firstColor--;
		}
		else
			t = mapping - firstColor;

		secondColor = firstColor + 1;

		for (i=0; i!=3; i++)
			mesh->m_colorArray.push_back(*(colorTable + 3*firstColor + i)*(1.0f - t) + *(colorTable + 3*secondColor + i)*t);
	}

	return true;
}

/**
*/
bool GenerateIsosurface(float isoLevel)
{
	CTimer timer;

	if (!m_polygonizer.get())
	{
		IMarchingMethod *polygonizer;

		if (m_isUsingCombinedTransforms)
			polygonizer = new CCombinedTransforms;
		else
			polygonizer = new CMarchingCubes;

		if (!polygonizer)
		{
			MarkError();

			return false;
		}

		if (!polygonizer->Load(m_image))
		{
			MarkError();

			return false;
		}

		if (!polygonizer->SetGridSize(m_image->GetWidth(), m_image->GetHeight(), m_image->GetDepth()))
		{
			MarkError();

			return false;
		}

		if (!polygonizer->SetTransformationStepCount(m_transformationStepCount))
		{
			MarkError();

			return false;
		}

		m_polygonizer.reset(polygonizer);

		if (!m_polygonizer.get())
		{
			MarkError();

			return false;
		}

		m_qualityMeasurer.reset(new CMeshQuality);

		if (!m_qualityMeasurer.get())
		{
			MarkError();

			return false;
		}
	}

	cout << "Generating isosurface at level " << isoLevel << "... ";

	timer.Restart();

	if (!m_polygonizer->GenerateMesh(isoLevel))
	{
		MarkError();

		return false;
	}

	cout << "ok (done in " << timer.GetElapsed() << " ms)." << endl;

	cout << "calculating bounding box...";

	m_meshBoundingBox.SetMesh(m_polygonizer->GetMesh());

	cout << "ok." << endl;

	// Outputs the algebraic error.
	cout << "Algebraic error: " << m_polygonizer->GetAlgebraicError() << "." << endl;

	m_qualityMeasurer->Load(m_polygonizer->GetMesh());

	m_qualityMeasurer->MeasureQuality();

	MapQualityToColorArray(m_polygonizer->GetMesh(), m_qualityMeasurer);

	cout << "Min/max/avg quality: " << m_qualityMeasurer->GetMinQuality() << "/" << m_qualityMeasurer->GetMaxQuality() << "/" << m_qualityMeasurer->GetAverageQuality() << "." << endl;

	return true;
}

/**
*/
bool SaveIsosurface(void)
{
	string fileName;

	fileName = m_imageName;

	/*switch (m_polygonizer->GetTransform())
	{
	case IMarchingMethod::NONE:
		break;
	case IMarchingMethod::GRADIENT_TRANSFORM:
		fileName += "_gradientTransform";
		break;
	case IMarchingMethod::TANGENTIAL_TRANSFORM:
		fileName += "_tangentTransform";
		break;
	}*/

	if (m_polygonizer->GetMesh()->IsValid())
		if (m_polygonizer->GetMesh()->Save(fileName, CVertexPool::OFF))
		{
			cout << fileName << ".off successfuly saved." << endl;

			return true;
		}

	MarkError();

	return false;
}

/**
*/
void DrawGrid(void)
{
    if (glIsList(m_gridDisplayList) == GL_TRUE)
        glCallList(m_gridDisplayList);
    else
    {
        m_gridDisplayList = glGenLists(1);

        glNewList(m_gridDisplayList, GL_COMPILE_AND_EXECUTE);

        glLineStipple(1, 0xaaaa);

        glEnable(GL_LINE_STIPPLE);

        glColor3f(0.5f, 0.5f, 0.5f);

        glBegin(GL_LINES);

        for (unsigned int i=0; i!=11; i++)
        {
            // yz
            glVertex3i( 0,  i,  0);
            glVertex3i( 0,  i, 10);

            glVertex3i( 0,  0,  i);
            glVertex3i( 0, 10,  i);

            // xz
            glVertex3i( i,  0,  0);
            glVertex3i( i,  0, 10);

            glVertex3i( 0,  0,  i);
            glVertex3i(10,  0,  i);

            // xy
            glVertex3i( i,  0,  0);
            glVertex3i( i, 10,  0);

            glVertex3i( 0,  i,  0);
            glVertex3i(10,  i,  0);
        }

        glEnd();

        glDisable(GL_LINE_STIPPLE);

        glLineWidth(2.0f);

        glBegin(GL_LINES);

        glColor3f(1.0f, 0.0f, 0.0f);
        glVertex3i( 0,  0,  0);
        glVertex3i(10,  0,  0);

        glColor3f(0.0f, 1.0f, 0.0f);
        glVertex3i( 0,  0,  0);
        glVertex3i( 0, 10,  0);

        glColor3f(0.0f, 0.0f, 1.0f);
        glVertex3i( 0,  0,  0);
        glVertex3i( 0,  0, 10);

        glEnd();

        glLineWidth(1.0f);

        glEndList();
    }
}

/**
*/
void DrawHelp(void)
{
	cout << endl;
	cout << "Available commands:" << endl;

	cout << "ESC:\t" << "Exit program." << endl;
	cout << "'e':\t" << "Enable/disable mesh edges drawing." << endl;
	cout << "'h':\t" << "Draw this help." << endl;
	cout << endl;
}

/**
*/
void EnableLight(void)
{
	float lightPosition[] =	{0.0f, 0.0f, -100.0f, 0.0f}, 
		lightDiffuse[] = {0.6f, 0.6f, 0.6f, 1.0f};

    glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
    
	glMatrixMode(GL_MODELVIEW);

	glPushMatrix();
	
	glLoadIdentity();

	glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);

	glPopMatrix();

	glLightfv(GL_LIGHT0, GL_DIFFUSE, lightDiffuse);

	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
}

/**
*/
void DisableLight(void)
{
	glDisable(GL_LIGHTING);
}

/**
*/
void OnDestroy(void)
{
	int windowName;

	windowName = glutGetWindow();

	if (windowName)
	{
		if (glIsList(m_gridDisplayList) == GL_TRUE)
			glDeleteLists(m_gridDisplayList, 1);

		glutDestroyWindow(windowName);
	}

	if (!CErrorManager::Instance().IsOk())
		CErrorManager::Instance().FlushErrors();

	exit(EXIT_SUCCESS);
}

/**
*/
void OnPaint(void)
{
	float width,
		height,
		depth,
		sizeMax;

    glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

    glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT);

    m_camera.ApplyTransform();

    glEnable(GL_DEPTH_TEST);
	
	DrawGrid();

	if (m_polygonizer->IsValid())
	{
		width = m_meshBoundingBox.m_xMax - m_meshBoundingBox.m_xMin;
		height = m_meshBoundingBox.m_yMax - m_meshBoundingBox.m_yMin;
		depth = m_meshBoundingBox.m_zMax - m_meshBoundingBox.m_zMin;

		sizeMax = max(width, max(height, depth));

		glScalef(10.0f/sizeMax, 10.0f/sizeMax, 10.0f/sizeMax);

		glTranslatef(-0.5f*(m_meshBoundingBox.m_xMax + m_meshBoundingBox.m_xMin), -0.5f*(m_meshBoundingBox.m_yMax + m_meshBoundingBox.m_yMin), -0.5f*(m_meshBoundingBox.m_zMax + m_meshBoundingBox.m_zMin));

		EnableLight();

		glEnable(GL_NORMALIZE);

		glEnable(GL_COLOR_MATERIAL);

        glPolygonOffset(1.0, 1.0);

		glEnable(GL_POLYGON_OFFSET_FILL);
		
		glColor3f(1.0f, 1.0f, 0.0f);
            
		m_polygonizer->GetMesh()->Draw();
        
        glDisable(GL_POLYGON_OFFSET_FILL);

		glDisable(GL_COLOR_MATERIAL);

		glDisable(GL_NORMALIZE);

		DisableLight();

        if (m_drawEdges)
        {
            glColor3f(0.3f, 0.3f, 0.3f);
            
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
            
			glVertexPointer(m_polygonizer->GetMesh()->GetVertexSize(), GL_FLOAT, 0, m_polygonizer->GetMesh()->GetVertexArray());

			glEnableClientState(GL_VERTEX_ARRAY);

			glDrawElements(m_polygonizer->GetMesh()->GetPrimitiveType(), m_polygonizer->GetMesh()->GetElementCount(), GL_UNSIGNED_INT, m_polygonizer->GetMesh()->GetElementArray());

			glDisableClientState(GL_VERTEX_ARRAY);

            glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        }
	}
  
    glDisable(GL_DEPTH_TEST);
	
    glutSwapBuffers();
}

/**
*/
void OnSize(GLsizei w, GLsizei h)
{
    if (h == 0)
        h = 1;
    
    glViewport(0, 0, w, h);

    m_camera.Create(45.0f, 1.0f, 50.0f, w, h);
}

/**
*/
void OnIdle(void)
{
    //glutPostRedisplay();
}

/**
*/
void OnVisible(int isVisible)
{
    if (isVisible == GLUT_VISIBLE)	
        glutIdleFunc(OnIdle);
    else 
        glutIdleFunc(0);
}

/**
*/
void OnKeyDown(unsigned char key, int x, int y) 
{
    switch (key)
    {
	case 'e':
	case 'E':
		m_drawEdges = !m_drawEdges;
		break;
	case 'h':
	case 'H':
		DrawHelp();
		break;
    case 'q':
    case 'Q':
    case 27:
        OnDestroy();
        
        exit(EXIT_SUCCESS);
        break;
	case 's':
	case 'S':
		if (!SaveIsosurface())
		{
			MarkError();

			OnDestroy();
		}
		break;
    }
    
    glutPostRedisplay();
}

/**
*/
void OnMouse(int button, int state, int x, int y)
{
    if (state == GLUT_DOWN)
        m_mouseButton = button;
    else
        m_mouseButton = 0;

    m_mouseX = x;
    m_mouseY = y;
}

/**
*/
void OnMouseMove(int x, int y)
{
    switch (m_mouseButton)
    {
    case GLUT_LEFT_BUTTON:
        m_camera.Pitch(0.1f*(y - m_mouseY));
        m_camera.Yaw(0.1f*(x - m_mouseX));
        break;

    case GLUT_MIDDLE_BUTTON:
        m_camera.MoveSide(0.025*(x - m_mouseX));
        m_camera.MoveUp(0.025f*(m_mouseY - y));
        break;

    case GLUT_RIGHT_BUTTON:
        m_camera.MoveFront(0.05f*(y - m_mouseY));
        break;
    }

    m_mouseX = x;
    m_mouseY = y;

    glutPostRedisplay();
}

/**
*/
void OnCreate(int argc, char** argv)
{
    glutInit(&argc, argv);
    glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGBA|GLUT_DEPTH);
    glutInitWindowSize(512, 512);

    glutCreateWindow("Test CSimpleCombinedTransforms implementation.");

    m_camera.Create(45.0f, 1.0f, 50.0f, glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT));

    m_camera.MoveFront(-20.0f);
    m_camera.ApplyTransform();  

    glutDisplayFunc(OnPaint);
    glutKeyboardFunc(OnKeyDown);
    glutMouseFunc(OnMouse);
    glutMotionFunc(OnMouseMove);
    glutReshapeFunc(OnSize);
    glutVisibilityFunc(OnVisible);
}

/**
*/
int main(int argc, char **argv)
{
	if (argc < 3)
	{
		cout << endl;
		cout << "SYPNOSIS" << endl;
		cout << "TestCombinedTransforms input(.nhdr) isoLevel output [-isCombined] [-transformSteps X]" << endl;
		cout << endl;

		OnDestroy();
	}

	if (!ParseCommandLineOptions(argc, argv))
	{
		MarkError();

		OnDestroy();
	}

	DrawHelp();

	if (!LoadImage(argv[1]))
	{
		MarkError();

		OnDestroy();
	}

	m_isoLevel = atof(argv[2]);

	if (!GenerateIsosurface(m_isoLevel))
	{
		MarkError();

		OnDestroy();
	}

#ifdef NON_INTERACTIVE
	if (!m_polygonizer->GetMesh()->Save(argv[3], CVertexPool::OFF))
	{
		MarkError();

		OnDestroy();
	}
#else
	OnCreate(argc, argv);
    
    glutMainLoop();
#endif
    
    return EXIT_SUCCESS;
}

