#ifndef GAGEADAPTOR_INCLUDED
#define GAGEADAPTOR_INCLUDED

#include <string>

#include <teem/nrrd.h>
#include <teem/gage.h>

#include <boost/utility.hpp>
#include <boost/shared_ptr.hpp>

#if TEEM_VERSION_MINOR >= 10
#define NRRD_FLOAT_TYPE double
#else
#define NRRD_FLOAT_TYPE float
#endif

class CGageAdaptor
	: boost::noncopyable
{
public:
	enum QUERY_ITEM {
		// Data value.
		VALUE = gageSclValue,
		// Gradient vector, normalized.
		NORMAL = gageSclNormal,
		// Gradient vector, un-normalized.
		GRADIENT = gageSclGradVec,
		// Gradient magnitude.
		GRADIENT_MAGNITUDE = gageSclGradMag,
		// Hessian (column-order).
		HESSIAN = gageSclHessian,
		// Laplacian: Dxx + Dyy + Dzz.
		LAPLACIAN = gageSclLaplacian,
		// Hessian's 1st eigenvalue.
		HESSIAN_1ST_EIGENVALUE = gageSclHessEval0,
		// Hessian's 2nd eigenvalue.
		HESSIAN_2ND_EIGENVALUE = gageSclHessEval1,
		// Hessian's 3rd eigenvalue.
		HESSIAN_3RD_EIGENVALUE = gageSclHessEval2,
		// 1st principle curvature.
		PRINCIPAL_CURVATURE = gageSclK1
	};
	enum VALUE_TYPE {
		// Signifies "type is unset/unknown".
		UNKNOWN_TYPE = nrrdTypeUnknown,
		// Signed 1-byte integer.
		BYTE = nrrdTypeChar,
		// Unsigned 1-byte integer.
		UNSIGNED_BYTE = nrrdTypeUChar,
		// Signed 2-byte integer.
		SHORT = nrrdTypeShort,
		// Unsigned 2-byte integer.
		UNSIGNED_SHORT = nrrdTypeUShort,
		// Signed 4-byte integer.
		INT = nrrdTypeInt,
		// Unsigned 4-byte integer.
		UNSIGNED_INT = nrrdTypeUInt,
		// 4-byte floating point.
		FLOAT = nrrdTypeFloat,
		// 8-byte floating point.
		DOUBLE = nrrdTypeDouble
	};
	CGageAdaptor(void);
	CGageAdaptor(const std::string& path);
	virtual ~CGageAdaptor(void);
	bool Open(const std::string& path);
	bool OpenFromMemory(void *data, VALUE_TYPE type, unsigned int width, unsigned int height, unsigned int depth);
	void Close(void);
	bool IsOpen(void) const;
	bool EnableQuery(int item);
	bool ResetKernel(void);
	void SetClamp(bool doClamp);
	bool SetValueKernel(const NrrdKernel *type, const double *parameters);
	bool Set1stDerivativeKernel(const NrrdKernel *type, const double *parameters);
	bool Set2ndDerivativeKernel(const NrrdKernel *type, const double *parameters);
	int GetWidth(void) const;
	int GetHeight(void) const;
	int GetDepth(void) const;
	VALUE_TYPE GetType(void) const;
	const NRRD_FLOAT_TYPE *GetValueArray(void) const;
	float GetValue(float x, float y, float z) const;
	const NRRD_FLOAT_TYPE *GetNormal(float x, float y, float z) const;
	const NRRD_FLOAT_TYPE *GetGradient(float x, float y, float z) const;
	float GetGradientMagnitude(float x, float y, float z) const;
	const NRRD_FLOAT_TYPE *GetHessian(float x, float y, float z) const;
	float GetLaplacian(float x, float y, float z) const;
	float GetHessian1stEigenvalue(float x, float y, float z) const;
	float GetHessian2ndEigenvalue(float x, float y, float z) const;
	float GetHessian3rdEigenvalue(float x, float y, float z) const;
	float Get1stPrincipalCurvature(float x, float y, float z) const;
private:
	bool OpenNrrd(const std::string& path);
	bool OpenNrrdFromMemory(void *data, VALUE_TYPE type, unsigned int width, unsigned int height, unsigned int depth);
	bool CreateDefaultContext(void);
	bool UpdateKernel(void);
	inline void Clamp(float *x, float *y, float *z) const;
	void Create(void);
protected:
	Nrrd *m_imageHandle;
	gageContext *m_measurementContext;
	gagePerVolume *m_imageInfo;
	const NRRD_FLOAT_TYPE *m_valuePointer;
	const NRRD_FLOAT_TYPE *m_normalPointer;
	const NRRD_FLOAT_TYPE *m_gradientPointer;
	const NRRD_FLOAT_TYPE *m_gradientMagnitudePointer;
	const NRRD_FLOAT_TYPE *m_hessianPointer;
	const NRRD_FLOAT_TYPE *m_laplacianPointer;
	const NRRD_FLOAT_TYPE *m_hessian1stEigenvaluePointer;
	const NRRD_FLOAT_TYPE *m_hessian2ndEigenvaluePointer;
	const NRRD_FLOAT_TYPE *m_hessian3rdEigenvaluePointer;
	const NRRD_FLOAT_TYPE *m_1stPrincipleCurvaturePointer;
	bool m_isOpen;
	bool m_isCopy;
	bool m_doClamp;
};

#endif // GAGEADAPTOR_INCLUDED

